import com.imsl.datamining.neural.*;
import java.io.*;

//*****************************************************************************
// Two Layer Feed-Forward Network with 4 inputs: 1 categorical with 3 classes
// encoded using binary encoding and 1 continuous input, and 1 output 
// target (continuous).  There is a perfect linear relationship between 
// the input and output variables:
//
// MODEL:  Y = 10*X1 + 20*X2 + 30*X3 + 2*X4 
//     
// Variables X1-X3 are the binary encoded nominal variable and X4 is the
// continuous variable.
//
// This example uses Linear Activation in both the hidden and output layers
// The network uses a 2-layer configuration, one hidden layer and one 
// output layer.  The hidden layer consists of 3 perceptrons.  The output 
// layer consists of a single output perceptron.
// The input from the continuous variable is scaled to [0,1] before training
// the network.  Training is done using the Quasi-Newton Trainer.
// The network has a total of 19 weights.  
// Since the network target is a linear combination of the network inputs, and
// since all perceptrons use linear activation, the network is able to forecast
// the every training target exactly.  The largest residual is 2.78E-08.
//*****************************************************************************

public class NetworkEx1 implements Serializable {
 // **********************************************************************
 // MAIN
 // **********************************************************************
    public static void main(String[] args) throws Exception {
       double xData[][]; // Input  Attributes for Training Patterns
       double yData[][]; // Output Attributes for Training Patterns
       double weight[];  // network weights
       double gradient[];// network gradient after training
       // Input Attributes for Forecasting
       double x[][] = { {1,0,0,0.0}, {0,1,0,0.0}, {0,0,1,0.0},
                        {1,0,0,5.0}, {0,1,0,5.0}, {0,0,1,5.0},
                        {1,0,0,10.0}, {0,1,0,10.0}, {0,0,1,10.0}
                      };
       double xTemp[], y[];// Temporary areas for storing forecasts
       int i, j;           // loop counters
       // Names of Serialized Files
       String networkFileName = "FeedForwardNetworkEx1.ser"; // the network
       String trainerFileName = "FeedForwardTrainerEx1.ser"; // the trainer
       String xDataFileName   = "FeedForwardxDataEx1.ser";   // xData
       String yDataFileName   = "FeedForwardyDataEx1.ser";   // yData
    // **********************************************************************
    // READ THE TRAINED NETWORK FROM THE SERIALIZED NETWORK OBJECT
    // **********************************************************************
       System.out.println("--> Reading Trained Network from " +
                               networkFileName);
       Network network = (Network)read(networkFileName);
    // **********************************************************************
    // READ THE SERIALIZED XDATA[][] AND YDATA[][] ARRAYS OF TRAINING
    // PATTERNS.
    // **********************************************************************
       System.out.println("--> Reading xData from " +
                               xDataFileName);
       xData = (double[][])read(xDataFileName);
       System.out.println("--> Reading yData from " +
                               yDataFileName);
       yData = (double[][])read(yDataFileName);
    // **********************************************************************
    // READ THE SERIALIZED TRAINER OBJECT
    // **********************************************************************
       System.out.println("--> Reading Network Trainer from " +
                               trainerFileName);
       Trainer trainer = (Trainer)read(trainerFileName);
    // **********************************************************************
    // DISPLAY TRAINING STATISTICS
    // **********************************************************************
       double stats[] = network.computeStatistics(xData, yData);
       // Display Network Errors
       System.out.println("***********************************************");
       System.out.println("--> SSE:                       "+(float)stats[0]);
       System.out.println("--> RMS:                       "+(float)stats[1]);
       System.out.println("--> Laplacian Error:           "+(float)stats[2]);
       System.out.println("--> Scaled Laplacian Error:    "+(float)stats[3]);
       System.out.println("--> Largest Absolute Residual: "+(float)stats[4]);
       System.out.println("***********************************************");
       System.out.println("");
       // **********************************************************************
       // OBTAIN AND DISPLAY NETWORK WEIGHTS AND GRADIENTS
       // **********************************************************************
       System.out.println("--> Getting Network Information");
       // Get weights               
       weight   = network.getWeights();
       // Get number of weights = number of gradients
       int nWeights = network.getNumberOfWeights();
       // Obtain Gradient Vector
       gradient = trainer.getErrorGradient();
       // Print Network Weights and Gradients
       System.out.println(" ");
       System.out.println("--> Network Weights and Gradients:");
       for(i=0; i < nWeights; i++){
        System.out.println("w["+i+"]=" + (float)weight[i]+
                          " g["+i+"]="+(float)gradient[i]);
       }
    // **********************************************************************
    // OBTAIN AND DISPLAY FORECASTS FOR THE LAST 10 TRAINING TARGETS
    // **********************************************************************
       // Get number of network inputs
       int nInputs  = network.getNumberOfInputs();
       // Get number of network outputs
       int nOutputs = network.getNumberOfOutputs();
       xTemp = new double[nInputs]; // temporary x space for forecast inputs
       y     = new double[nOutputs];// temporary y space for forecast output
       System.out.println(" ");
       // Obtain example forecasts for input attributes = x[]
       // X1-X3 are binary encoded for one nominal variable with 3 classes
       // X4 is a continuous input attribute ranging from 0-10.  During
       // training, X4 was scaled to [0,1] by dividing by 10.
       for(i=0;i<9;i++){
          for(j=0;j<nInputs;j++) xTemp[j] = x[i][j];
          xTemp[nInputs-1] = xTemp[nInputs-1]/10.0;
          y = network.forecast(xTemp);
          System.out.print("--> X1="+(int)x[i][0]+
                             " X2="+(int)x[i][1]+" X3="+(int)x[i][2]+
                             " | X4="+x[i][3]);
          System.out.println(" | y="+
               (float)(10.0*x[i][0]+20.0*x[i][1]+30.0*x[i][2]+2.0*x[i][3])+
               "| Forecast="+(float)y[0]);
       }
    }
// **************************************************************************
// READ SERIALIZED NETWORK FROM A FILE
// **************************************************************************
    static public Object read(String filename) 
       throws IOException, ClassNotFoundException {
       FileInputStream fis = new FileInputStream(filename);
       ObjectInputStream ois = new ObjectInputStream(fis);
       Object obj = ois.readObject();
       ois.close();
       fis.close();
       return obj;
    }
}
